@@ -9,9 +9,13 @@ use clap::{
 	builder::TypedValueParser, error::ErrorKind, Arg, ArgAction, Command, CommandFactory, Parser,
 	ValueEnum, ValueHint,
 };
+use miette::{IntoDiagnostic, Result};
+use tokio::{fs::File, io::AsyncReadExt};
 use watchexec::paths::PATH_SEPARATOR;
 use watchexec_signals::Signal;
 
+use crate::filterer::parse::parse_filter_program;
+
 const OPTSET_FILTERING: &str = "Filtering";
 const OPTSET_COMMAND: &str = "Command";
 const OPTSET_DEBUGGING: &str = "Debugging";
@@ -155,30 +159,19 @@ pub struct Args {
 	#[arg(
 		short,
 		long,
-		default_value = "queue",
+		default_value = "do-nothing",
 		hide_default_value = true,
 		value_name = "MODE"
 	)]
 	pub on_busy_update: OnBusyUpdate,
 
-	/// Deprecated alias for '--on-busy-update=do-nothing'
-	///
-	/// This option is deprecated and will be removed in the next major release.
-	#[arg(
-		long,
-		short = 'W',
-		hide = true,
-		conflicts_with_all = ["on_busy_update", "restart"],
-	)]
-	pub watch_when_idle: bool,
-
 	/// Restart the process if it's still running
 	///
 	/// This is a shorthand for '--on-busy-update=restart'.
 	#[arg(
 		short,
 		long,
-		conflicts_with_all = ["on_busy_update", "watch_when_idle"],
+		conflicts_with_all = ["on_busy_update"],
 	)]
 	pub restart: bool,
 
@@ -195,15 +188,11 @@ pub struct Args {
 	#[arg(
 		short,
 		long,
-		conflicts_with_all = ["restart", "watch_when_idle"],
+		conflicts_with_all = ["restart"],
 		value_name = "SIGNAL"
 	)]
 	pub signal: Option<Signal>,
 
-	/// Hidden legacy shorthand for '--signal=kill'.
-	#[arg(short, long, hide = true)]
-	pub kill: bool,
-
 	/// Signal to send to stop the command
 	///
 	/// This is used by 'restart' and 'signal' modes of '--on-busy-update' (unless '--signal' is
@@ -230,14 +219,15 @@ pub struct Args {
 	/// it is forcefully terminated.
 	///
 	/// Takes a unit-less value in seconds, or a time span value such as "5min 20s".
+	/// Providing a unit-less value is deprecated and will warn; it will be an error in the future.
 	///
-	/// The default is 60 seconds. Set to 0 to immediately force-kill the command.
+	/// The default is 10 seconds. Set to 0 to immediately force-kill the command.
 	///
 	/// This has no practical effect on Windows as the command is always forcefully terminated; see
 	/// '--stop-signal' for why.
 	#[arg(
 		long,
-		default_value = "60",
+		default_value = "10s",
 		hide_default_value = true,
 		value_name = "TIMEOUT"
 	)]
@@ -277,12 +267,13 @@ pub struct Args {
 	/// every accumulated event will build up in memory.
 	///
 	/// Takes a unit-less value in milliseconds, or a time span value such as "5sec 20ms".
+	/// Providing a unit-less value is deprecated and will warn; it will be an error in the future.
 	///
 	/// The default is 50 milliseconds. Setting to 0 is highly discouraged.
 	#[arg(
 		long,
 		short,
-		default_value = "50",
+		default_value = "50ms",
 		hide_default_value = true,
 		value_name = "TIMEOUT"
 	)]
@@ -327,14 +318,10 @@ pub struct Args {
 	/// VCS ignore files (Git, Mercurial, Bazaar, Darcs, Fossil) are only used if the corresponding
 	/// VCS is discovered to be in use for the project/origin. For example, a .bzrignore in a Git
 	/// repository will be discarded.
-	///
-	/// Note that this was previously called '--no-ignore', but that's now deprecated and its use is
-	/// discouraged, as it may be repurposed in the future.
 	#[arg(
 		long,
 		help_heading = OPTSET_FILTERING,
 		verbatim_doc_comment,
-		alias = "no-ignore", // deprecated
 	)]
 	pub no_project_ignore: bool,
 
@@ -409,6 +396,7 @@ pub struct Args {
 	/// but portable and slightly more efficient.
 	///
 	/// Takes a unit-less value in seconds, or a time span value such as "2min 5s".
+	/// Providing a unit-less value is deprecated and will warn; it will be an error in the future.
 	#[arg(long, value_name = "DURATION")]
 	pub delay_run: Option<TimeSpan>,
 
@@ -421,6 +409,7 @@ pub struct Args {
 	///
 	/// Optionally takes a unit-less value in milliseconds, or a time span value such as "2s 500ms",
 	/// to use as the polling interval. If not specified, the default is 30 seconds.
+	/// Providing a unit-less value is deprecated and will warn; it will be an error in the future.
 	///
 	/// Aliased as '--force-poll'.
 	#[arg(
@@ -434,18 +423,18 @@ pub struct Args {
 
 	/// Use a different shell
 	///
-	/// By default, Watchexec will use 'sh' on unix and 'cmd' (CMD.EXE) on Windows. With this, you
-	/// can override that and use a different shell, for example one with more features or one which
-	/// has your custom aliases and functions.
+	/// By default, Watchexec will use '$SHELL' if it's defined or a default of 'sh' on Unix-likes,
+	/// and either 'pwsh', 'powershell', or 'cmd' (CMD.EXE) on Windows, depending on what Watchexec
+	/// detects is the running shell.
+	///
+	/// With this option, you can override that and use a different shell, for example one with more
+	/// features or one which has your custom aliases and functions.
 	///
 	/// If the value has spaces, it is parsed as a command line, and the first word used as the
 	/// shell program, with the rest as arguments to the shell.
 	///
 	/// The command is run with the '-c' flag (except for 'cmd' on Windows, where it's '/C').
 	///
-	/// Note that the default shell will change at the next major release: the value of '$SHELL'
-	/// will be respected, falling back to 'sh' on unix and to PowerShell on Windows.
-	///
 	/// The special value 'none' can be used to disable shell use entirely. In that case, the
 	/// command provided to Watchexec will be parsed, with the first word being the executable and
 	/// the rest being the arguments, and executed directly. Note that this parsing is rudimentary,
@@ -465,7 +454,7 @@ pub struct Args {
 	///
 	///   $ watchexec --shell=pwsh -- Test-Connection localhost
 	///
-	/// Use with cmd (default on Windows):
+	/// Use with CMD.exe:
 	///
 	///   $ watchexec --shell=cmd -- dir
 	///
@@ -483,41 +472,28 @@ pub struct Args {
 	)]
 	pub shell: Option<String>,
 
-	/// Don't use a shell
-	///
-	/// This is a shorthand for '--shell=none'.
+	/// Shorthand for '--shell=none'
 	#[arg(
 		short = 'n',
 		help_heading = OPTSET_COMMAND,
 	)]
 	pub no_shell: bool,
 
-	/// Don't use a shell
-	///
-	/// This is a deprecated alias for '--shell=none'.
-	#[arg(
-		long,
-		hide = true,
-		help_heading = OPTSET_COMMAND,
-		alias = "no-shell", // deprecated
-	)]
-	pub no_shell_long: bool,
-
-	/// Shorthand for '--emit-events=none'
+	/// Deprecated shorthand for '--emit-events=none'
 	///
 	/// This is the old way to disable event emission into the environment. See '--emit-events' for
-	/// more.
+	/// more. Will be removed at next major release.
 	#[arg(
 		long,
 		help_heading = OPTSET_COMMAND,
-		// TODO: deprecate then remove
+		hide = true, // deprecated
 	)]
 	pub no_environment: bool,
 
 	/// Configure event emission
 	///
-	/// Watchexec emits event information when running a command, which can be used by the command
-	/// to target specific changed files.
+	/// Watchexec can emit event information when running a command, which can be used by the child
+	/// process to target specific changed files.
 	///
 	/// One thing to take care with is assuming inherent behaviour where there is only chance.
 	/// Notably, it could appear as if the `RENAMED` variable contains both the original and the new
@@ -528,32 +504,14 @@ pub struct Args {
 	/// whether it was the old or new isn't known), rename events might split across two debouncing
 	/// boundaries, and so on.
 	///
-	/// This option controls where that information is emitted. It defaults to 'environment', which
-	/// sets environment variables with the paths of the affected files, for filesystem events:
+	/// This option controls where that information is emitted. It defaults to 'none', which doesn't
+	/// emit event information at all. The other options are 'environment' (deprecated), 'stdio',
+	/// 'file', 'json-stdio', and 'json-file'.
 	///
-	/// $WATCHEXEC_COMMON_PATH is set to the longest common path of all of the below variables,
-	/// and so should be prepended to each path to obtain the full/real path. Then:
-	///
-	///   - $WATCHEXEC_CREATED_PATH is set when files/folders were created
-	///   - $WATCHEXEC_REMOVED_PATH is set when files/folders were removed
-	///   - $WATCHEXEC_RENAMED_PATH is set when files/folders were renamed
-	///   - $WATCHEXEC_WRITTEN_PATH is set when files/folders were modified
-	///   - $WATCHEXEC_META_CHANGED_PATH is set when files/folders' metadata were modified
-	///   - $WATCHEXEC_OTHERWISE_CHANGED_PATH is set for every other kind of pathed event
-	///
-	/// Multiple paths are separated by the system path separator, ';' on Windows and ':' on unix.
-	/// Within each variable, paths are deduplicated and sorted in binary order (i.e. neither
-	/// Unicode nor locale aware).
-	///
-	/// This is the legacy mode and will be deprecated and removed in the future. The environment of
-	/// a process is a very restricted space, while also limited in what it can usefully represent.
-	/// Large numbers of files will either cause the environment to be truncated, or may error or
-	/// crash the process entirely.
-	///
-	/// Two new modes are available: 'stdio' writes absolute paths to the stdin of the command,
-	/// one per line, each prefixed with `create:`, `remove:`, `rename:`, `modify:`, or `other:`,
-	/// then closes the handle; 'file' writes the same thing to a temporary file, and its path is
-	/// given with the $WATCHEXEC_EVENTS_FILE environment variable.
+	/// The 'stdio' and 'file' modes are text-based: 'stdio' writes absolute paths to the stdin of
+	/// the command, one per line, each prefixed with `create:`, `remove:`, `rename:`, `modify:`,
+	/// or `other:`, then closes the handle; 'file' writes the same thing to a temporary file, and
+	/// its path is given with the $WATCHEXEC_EVENTS_FILE environment variable.
 	///
 	/// There are also two JSON modes, which are based on JSON objects and can represent the full
 	/// set of events Watchexec handles. Here's an example of a folder being created on Linux:
@@ -610,13 +568,33 @@ pub struct Args {
 	/// events to it, and provide the path to the file with the $WATCHEXEC_EVENTS_FILE
 	/// environment variable.
 	///
-	/// Finally, the special 'none' mode will disable event emission entirely.
-	// TODO: when deprecating, make the none mode the default.
+	/// Finally, the 'environment' mode was the default until 2.0. It sets environment variables
+	/// with the paths of the affected files, for filesystem events:
+	///
+	/// $WATCHEXEC_COMMON_PATH is set to the longest common path of all of the below variables,
+	/// and so should be prepended to each path to obtain the full/real path. Then:
+	///
+	///   - $WATCHEXEC_CREATED_PATH is set when files/folders were created
+	///   - $WATCHEXEC_REMOVED_PATH is set when files/folders were removed
+	///   - $WATCHEXEC_RENAMED_PATH is set when files/folders were renamed
+	///   - $WATCHEXEC_WRITTEN_PATH is set when files/folders were modified
+	///   - $WATCHEXEC_META_CHANGED_PATH is set when files/folders' metadata were modified
+	///   - $WATCHEXEC_OTHERWISE_CHANGED_PATH is set for every other kind of pathed event
+	///
+	/// Multiple paths are separated by the system path separator, ';' on Windows and ':' on unix.
+	/// Within each variable, paths are deduplicated and sorted in binary order (i.e. neither
+	/// Unicode nor locale aware).
+	///
+	/// This is the legacy mode, is deprecated, and will be removed in the future. The environment
+	/// is a very restricted space, while also limited in what it can usefully represent. Large
+	/// numbers of files will either cause the environment to be truncated, or may error or crash
+	/// the process entirely. The $WATCHEXEC_COMMON_PATH is also unintuitive, as demonstrated by the
+	/// multiple confused queries that have landed in my inbox over the years.
 	#[arg(
 		long,
 		help_heading = OPTSET_COMMAND,
 		verbatim_doc_comment,
-		default_value = "environment",
+		default_value = "none",
 		hide_default_value = true,
 		value_name = "MODE",
 		required_if_eq("only_emit_events", "true"),
@@ -658,12 +636,31 @@ pub struct Args {
 	/// By default, Watchexec will run the command in a process group, so that signals and
 	/// terminations are sent to all processes in the group. Sometimes that's not what you want, and
 	/// you can disable the behaviour with this option.
+	///
+	/// Deprecated, use '--wrap-process=none' instead.
 	#[arg(
 		long,
 		help_heading = OPTSET_COMMAND,
 	)]
 	pub no_process_group: bool,
 
+	/// Configure how the process is wrapped
+	///
+	/// By default, Watchexec will run the command in a process group in Unix, and in a Job Object
+	/// in Windows.
+	///
+	/// Some Unix programs prefer running in a session, while others do not work in a process group.
+	///
+	/// Use 'group' to use a process group, 'session' to use a process session, and 'none' to run
+	/// the command directly. On Windows, either of 'group' or 'session' will use a Job Object.
+	#[arg(
+		long,
+		help_heading = OPTSET_COMMAND,
+		value_name = "MODE",
+		default_value = "group",
+	)]
+	pub wrap_process: WrapMode,
+
 	/// Testing only: exit Watchexec after the first run
 	#[arg(short = '1', hide = true)]
 	pub once: bool,
@@ -680,6 +677,8 @@ pub struct Args {
 	pub notify: bool,
 
 	/// When to use terminal colours
+	///
+	/// Setting the environment variable `NO_COLOR` to any value is equivalent to `--color=never`.
 	#[arg(
 		long,
 		help_heading = OPTSET_OUTPUT,
@@ -789,6 +788,77 @@ pub struct Args {
 	)]
 	pub filter_files: Vec<PathBuf>,
 
+	/// [experimental] Filter programs.
+	///
+	/// /!\ This option is EXPERIMENTAL and may change and/or vanish without notice.
+	///
+	/// Provide your own custom filter programs in jaq (similar to jq) syntax. Programs are given
+	/// an event in the same format as described in '--emit-events-to' and must return a boolean.
+	///
+	/// In addition to the jaq stdlib, watchexec adds some custom filter definitions:
+	///
+	///   - 'path | file_meta' returns file metadata or null if the file does not exist.
+	///
+	///   - 'path | file_size' returns the size of the file at path, or null if it does not exist.
+	///
+	///   - 'path | file_read(bytes)' returns a string with the first n bytes of the file at path.
+	///     If the file is smaller than n bytes, the whole file is returned. There is no filter to
+	///     read the whole file at once to encourage limiting the amount of data read and processed.
+	///
+	///   - 'string | hash', and 'path | file_hash' return the hash of the string or file at path.
+	///     No guarantee is made about the algorithm used: treat it as an opaque value.
+	///
+	///   - 'any | kv_store(key)', 'kv_fetch(key)', and 'kv_clear' provide a simple key-value store.
+	///     Data is kept in memory only, there is no persistence. Consistency is not guaranteed.
+	///
+	///   - 'any | printout', 'any | printerr', and 'any | log(level)' will print or log any given
+	///     value to stdout, stderr, or the log (levels = error, warn, info, debug, trace), and
+	///     pass the value through (so '[1] | log("debug") | .[]' will produce a '1' and log '[1]').
+	///
+	/// All filtering done with such programs, and especially those using kv or filesystem access,
+	/// is much slower than the other filtering methods. If filtering is too slow, events will back
+	/// up and stall watchexec. Take care when designing your filters.
+	///
+	/// If the argument to this option starts with an '@', the rest of the argument is taken to be
+	/// the path to a file containing a jaq program.
+	///
+	/// Jaq programs are run in order, after all other filters, and short-circuit: if a filter (jaq
+	/// or not) rejects an event, execution stops there, and no other filters are run. Additionally,
+	/// they stop after outputting the first value, so you'll want to use 'any' or 'all' when
+	/// iterating, otherwise only the first item will be processed, which can be quite confusing!
+	///
+	/// Find user-contributed programs or submit your own useful ones at
+	/// <https://github.com/watchexec/watchexec/discussions/592>.
+	///
+	/// ## Examples:
+	///
+	/// Regexp ignore filter on paths:
+	///
+	///   'all(.tags[] | select(.kind == "path"); .absolute | test("[.]test[.]js$")) | not'
+	///
+	/// Pass any event that creates a file:
+	///
+	///   'any(.tags[] | select(.kind == "fs"); .simple == "create")'
+	///
+	/// Pass events that touch executable files:
+	///
+	///   'any(.tags[] | select(.kind == "path" && .filetype == "file"); .absolute | metadata | .executable)'
+	///
+	/// Ignore files that start with shebangs:
+	///
+	///   'any(.tags[] | select(.kind == "path" && .filetype == "file"); .absolute | read(2) == "#!") | not'
+	#[arg(
+		long = "filter-prog",
+		short = 'j',
+		help_heading = OPTSET_FILTERING,
+		value_name = "EXPRESSION",
+	)]
+	pub filter_programs: Vec<String>,
+
+	#[doc(hidden)]
+	#[clap(skip)]
+	pub filter_programs_parsed: Vec<jaq_syn::Main>,
+
 	/// Filename patterns to filter out
 	///
 	/// Provide a glob-like filter pattern, and events for files matching the pattern will be
@@ -857,7 +927,6 @@ pub struct Args {
 	/// Use '-v' when you need more diagnostic information.
 	#[arg(
 		long,
-		alias = "changes-only", // deprecated
 		help_heading = OPTSET_DEBUGGING,
 	)]
 	pub print_events: bool,
@@ -933,10 +1002,8 @@ pub struct Args {
 pub enum EmitEvents {
 	#[default]
 	Environment,
-	#[value(alias("stdin"))]
 	Stdio,
 	File,
-	#[value(alias("json-stdin"))]
 	JsonStdio,
 	JsonFile,
 	None,
@@ -951,6 +1018,14 @@ pub enum OnBusyUpdate {
 	Signal,
 }
 
+#[derive(Clone, Copy, Debug, Default, ValueEnum)]
+pub enum WrapMode {
+	#[default]
+	Group,
+	Session,
+	None,
+}
+
 #[derive(Clone, Copy, Debug, Default, ValueEnum)]
 pub enum ClearMode {
 	#[default]
@@ -995,7 +1070,10 @@ impl<const UNITLESS_NANOS_MULTIPLIER: u64> FromStr for TimeSpan<UNITLESS_NANOS_M
 		s.parse::<u64>()
 			.map_or_else(
 				|_| humantime::parse_duration(s),
-				|unitless| Ok(Duration::from_nanos(unitless * UNITLESS_NANOS_MULTIPLIER)),
+				|unitless| {
+					eprintln!("Warning: unitless time span values are deprecated and will be removed in an upcoming version");
+					Ok(Duration::from_nanos(unitless * UNITLESS_NANOS_MULTIPLIER))
+				},
 			)
 			.map(TimeSpan)
 	}
@@ -1084,26 +1162,24 @@ fn expand_args_up_to_doubledash() -> Result<Vec<OsString>, std::io::Error> {
 }
 
 #[inline]
-pub fn get_args() -> Args {
-	use tracing::{debug, warn};
+pub async fn get_args() -> Result<Args> {
+	use tracing::{debug, trace, warn};
 
 	if std::env::var("RUST_LOG").is_ok() {
 		warn!("⚠ RUST_LOG environment variable set, logging options have no effect");
 	}
 
-	if let Ok(filt) = std::env::var("WATCHEXEC_FILTERER") {
-		warn!("WATCHEXEC_FILTERER is deprecated");
-		if filt == "tagged" {
-			eprintln!("Tagged filterer has been removed. Open an issue if you have no workaround.");
-		}
-	}
-
 	debug!("expanding @argfile arguments if any");
 	let args = expand_args_up_to_doubledash().expect("while expanding @argfile");
 
 	debug!("parsing arguments");
 	let mut args = Args::parse_from(args);
 
+	// https://no-color.org/
+	if args.color == ColourMode::Auto && std::env::var("NO_COLOR").is_ok() {
+		args.color = ColourMode::Never;
+	}
+
 	if args.ignore_nothing {
 		args.no_global_ignore = true;
 		args.no_vcs_ignore = true;
@@ -1112,22 +1188,20 @@ pub fn get_args() -> Args {
 		args.no_discover_ignore = true;
 	}
 
-	if args.kill {
-		args.signal = Some(Signal::ForceStop);
-	}
-
 	if args.signal.is_some() {
 		args.on_busy_update = OnBusyUpdate::Signal;
 	} else if args.restart {
 		args.on_busy_update = OnBusyUpdate::Restart;
-	} else if args.watch_when_idle {
-		args.on_busy_update = OnBusyUpdate::DoNothing;
 	}
 
 	if args.no_environment {
 		args.emit_events_to = EmitEvents::None;
 	}
 
+	if args.no_process_group {
+		args.wrap_process = WrapMode::None;
+	}
+
 	if args.filter_fs_meta {
 		args.filter_fs_events = vec![
 			FsEvent::Create,
@@ -1150,6 +1224,24 @@ pub fn get_args() -> Args {
 			.exit();
 	}
 
+	for (n, prog) in args.filter_programs.iter_mut().enumerate() {
+		if let Some(progpath) = prog.strip_prefix('@') {
+			trace!(?n, path=?progpath, "reading filter program from file");
+			let mut progfile = File::open(&progpath).await.into_diagnostic()?;
+			let mut buf =
+				String::with_capacity(progfile.metadata().await.into_diagnostic()?.len() as _);
+			let bytes_read = progfile.read_to_string(&mut buf).await.into_diagnostic()?;
+			debug!(?n, path=?progpath, %bytes_read, "read filter program from file");
+			*prog = buf;
+		}
+	}
+
+	args.filter_programs_parsed = std::mem::take(&mut args.filter_programs)
+		.into_iter()
+		.enumerate()
+		.map(parse_filter_program)
+		.collect::<Result<_, _>>()?;
+
 	debug!(?args, "got arguments");
-	args
+	Ok(args)
 }
